package bcn.generator

import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.emf.common.util.EList
import org.eclipse.xtext.generator.IGenerator
import org.eclipse.xtext.generator.IFileSystemAccess
import java.util.Iterator
import bcn.lang.Controller
import bcn.lang.BehaviorDef
import bcn.lang.Instance
import bcn.lang.Connection
import bcn.lang.Declaration
import bcn.lang.JavaParam
import bcn.lang.PortRef


class LangGenerator implements IGenerator {
	
	override void doGenerate(Resource resource, IFileSystemAccess fsa) {
		generate(resource, fsa);
	}
	
	
	def public void generate(Resource resource, IFileSystemAccess fsa) {
	    generate(resource.allContents.filter(typeof(Controller)).head, fsa);
	}
	
	
	def public void generate(Controller controller, IFileSystemAccess fsa) {
	    // generate behaviors
	    for (Declaration m: controller.decls) {
	        if (m instanceof BehaviorDef) {
	            generate(m as BehaviorDef, fsa, controller.pckg)
	        }
	    }
	    
	    // write stub file
        fsa.generateFile(controller.pckg + "_stub/" + controller.name + "_stub.java",
'''
package «controller.pckg»_stub;

import com.cyberbotics.webots.controller.Robot;
import java.util.Random;
import bcn.Scheduler;
import bcn.coord.*;
import «controller.pckg».*;


public abstract class «controller.name»_stub {
    
    protected Scheduler scheduler;
    «FOR i: controller.decls»
    «IF i instanceof Instance»
    protected «(i as Instance).behavior»«(i as Instance).types» «(i as Instance).name»;
    «ENDIF»
    «ENDFOR»
    
    
    public abstract void setupAndRun();
    
    
    protected void connectAndRun(Robot robot) {
        connectAndRun(robot, 32);
    }
    
    
    protected void connectAndRun(Robot robot, int timeStep) {
        connectAndRun(robot, timeStep, new Random(0));
    }
    
    
    protected void connectAndRun(Robot robot, int timeStep, Random rnd) {
        // connect behaviors
        «FOR c: controller.decls»
        «IF c instanceof Connection»
        «portRef((c as Connection).src)».connectTo(«listPortRefs((c as Connection).dests)»);
        «ENDIF»
        «ENDFOR»
        // run scheduler
        scheduler = new Scheduler(robot, timeStep, rnd«FOR i: controller.decls»«IF i instanceof Instance», «(i as Instance).name»«ENDIF»«ENDFOR»);
        scheduler.run();
    }

}
'''
        );
        
        // write user file
        fsa.generateFile(controller.pckg + "/" + controller.name + ".java",
'''
package «controller.pckg»;


public class «controller.name» extends «controller.pckg»_stub.«controller.name»_stub {
    
    static public void main(String[] args) {
        «controller.name» controller = new «controller.name»();
        controller.setupAndRun();
    }
    
    
    @Override
    public void setupAndRun() {
        com.cyberbotics.webots.controller.Robot robot;
        
        // init sensors
        //TODO
        
        // init actuators
        //TODO
        
        // setup behaviors
        «FOR d: controller.decls»
        «IF d instanceof Instance»
        «(d as Instance).name» = new «(d as Instance).behavior»(...);
        «ENDIF»
        «ENDFOR»
        
        // run controller
        connectAndRun(robot);
    }
    
}
'''
        );
	}
	
	
	def private String portRef(PortRef ref) {
	    return '''«ref.behavior.name».get_«ref.port»(«IF ref.index != null»«ref.index»«ENDIF»)''';
	}
    
    
    def private String listPortRefs(EList<PortRef> refs) {
        val Iterator<PortRef> i = refs.iterator;
        var PortRef p = i.next;
        val StringBuilder s = new StringBuilder();
        s.append(portRef(p));
        while (i.hasNext) {
            p = i.next;
            s.append(", ");
            s.append(portRef(p));
        }
        return s.toString;
    }
	
	
	def public void generate(BehaviorDef behavior, IFileSystemAccess fsa, String pckg) {
	    // write stub file
	    fsa.generateFile(pckg + "_stub/" + behavior.name + "_stub.java",
'''
package «pckg»_stub;

import bcn.*;
import «pckg».*;


public abstract class «behavior.name»_stub«behavior.types» extends Behavior {
    
    «FOR in: behavior.inputs»
    «IF in.port != null»
    protected final InputPort<«in.port.type»>«IF in.port.array != null»[]«ENDIF» _«in.port.name»;
    «ELSEIF in.influence != null»
    protected final InputInfluence«IF in.influence.array != null»[]«ENDIF» _«in.influence.name»;
    «ENDIF»
    «ENDFOR»
    «FOR out: behavior.outputs»
    protected final OutputPort<«out.type»>«IF out.array != null»[]«ENDIF» _«out.name»;
    «ENDFOR»
    
    public «behavior.name»_stub(«formalParams(behavior.params)») {
        «FOR in: behavior.inputs»
        «IF in.port != null»
        «IF in.port.array == null»
        _«in.port.name» = new «IF in.buffer == null»InputRegister<«in.port.type»>()«ELSE»InputBuffer<«in.port.type»>(«in.buffer»)«ENDIF»;
        «ELSE»
        _«in.port.name» = new «IF in.buffer == null»InputRegister«ELSE»InputBuffer«ENDIF»[«in.port.array»];
        for (int i=0; i<«in.port.array»; i++) {
            _«in.port.name»[i] = new «IF in.buffer == null»InputRegister<«in.port.type»>()«ELSE»InputBuffer<«in.port.type»>(«in.buffer»)«ENDIF»;
        }
        «ENDIF»
        «ELSEIF in.influence != null»
        «IF in.influence.array == null»
        _«in.influence.name» = new InputInfluence((float)(«in.influence.min»), (float)(«in.influence.max»), (float)(«in.influence.dec»));
        «ELSE»
        _«in.influence.name» = new InputInfluence[«in.influence.array»];
        for (int i=0; i<«in.influence.array»; i++) {
            _«in.influence.name»[i] = new InputInfluence((float)(«in.influence.min»), (float)(«in.influence.max»), (float)(«in.influence.dec»));
        }
        «ENDIF»
        «ENDIF»
        «ENDFOR»
        «FOR out: behavior.outputs»
        «IF out.array == null»
        _«out.name» = new OutputPort<«out.type»>();
        «ELSE»
        _«out.name» = new OutputPort[«out.array»];
        for (int i=0; i<«out.array»; i++) {
            _«out.name»[i] = new OutputPort<«out.type»>();
        }
        «ENDIF»
        «ENDFOR»
    }
    
    «FOR in: behavior.inputs»
    «IF in.port != null»
    public final InputPort<«in.port.type»> get_«in.port.name»(«IF in.port.array != null»int id«ENDIF») {
        return _«in.port.name»«IF in.port.array != null»[id]«ENDIF»;
    }
    
    «ELSEIF in.influence != null»
    public final InputInfluence get_«in.influence.name»(«IF in.influence.array != null»int id«ENDIF») {
        return _«in.influence.name»«IF in.influence.array != null»[id]«ENDIF»;
    }
    
    «ENDIF»
    «ENDFOR»
    «FOR out: behavior.outputs»
    public final OutputPort<«out.type»> get_«out.name»(«IF out.array != null»int id«ENDIF») {
        return _«out.name»«IF out.array != null»[id]«ENDIF»;
    }
    
    «ENDFOR»
}
'''
	    );
	    
	    // write user file
	    fsa.generateFile(pckg + "/" + behavior.name + ".java",
'''
package «pckg»;


public class «behavior.name»«behavior.types» extends «pckg»_stub.«behavior.name»_stub«behavior.types» {
    
    public «behavior.name»(«formalParams(behavior.params)») {
        super(«actualParams(behavior.params)»);
    }
    
    
    @Override
    public void step() {
        //TODO
    }
    
}
'''
	    );
	}
	
	
	def private String formalParams(EList<JavaParam> params) {
        return listParams(params, true);
	}
    
    
    def private String actualParams(EList<JavaParam> params) {
        return listParams(params, false);
    }
    
    
    def private String listParams(EList<JavaParam> params, boolean formal) {
        if (params.size < 1) {
            return "";
        } else {
            val Iterator<JavaParam> i = params.iterator;
            var JavaParam p = i.next;
            val StringBuilder s = new StringBuilder();
            if (formal) {
                s.append(p.type + ' ');
            }
            s.append(p.name);
            while (i.hasNext) {
                p = i.next;
                s.append(", ");
                if (formal) {
                    s.append(p.type + ' ');
                }
                s.append(p.name);
            }
            return s.toString;
        }
    }
	
}
