package bcn.coord;

import bcn.InputPort;
import bcn.InputRegister;
import bcn.Behavior;
import bcn.OutputPort;


public class ActivationCoord<T> extends Behavior {
    
    protected final InputPort<MsgAct<T>>[] _in;
    protected final OutputPort<MsgAct<T>> _out;
    protected final OutputPort<T> _outMsg;
    protected final OutputPort<Float> _outAct;
    protected final OutputPort<Integer> _winner;
    
    
    public ActivationCoord(int count) {
    	if (count < 2) {
    		throw new IllegalArgumentException("A coordinator with less than 2 inputs doesn't make sense!");
    	}
        _in = new InputPort[count];
        for (int i=0; i<count; i++) {
            _in[i] = new InputRegister<MsgAct<T>>();
            _in[i].put(MsgAct.ACT_0);  // fill inputs with dummy data
        }
        _out = new OutputPort<MsgAct<T>>();
        _outMsg = new OutputPort<T>();
        _outAct = new OutputPort<Float>();
        _winner = new OutputPort<Integer>();
    }
    
    
    public final InputPort<MsgAct<T>> get_in(int id) {
        return _in[id];
    }
    
    
    public final OutputPort<MsgAct<T>> get_out() {
        return _out;
    }
    
    
    public final OutputPort<T> get_outMsg() {
        return _outMsg;
    }
    
    
    public final OutputPort<Float> get_outAct() {
        return _outAct;
    }
    
    
    public final OutputPort<Integer> get_winner() {
        return _winner;
    }
    
    
    @Override
    public void step() {
    	/* The message with the highest activation level takes control:
    	 * messages are not removed from the input registers;
    	 * if a behavior doesn't need to take control it sends messages with activation equal to 0.*/
    	MsgAct<T> winner = _in[0].peek();  // first input (could be dummy data)
        int id = 0;
        for (int i=1; i<_in.length; i++) {
        	MsgAct<T> msg = _in[i].peek();
            if (msg.act > winner.act) {
                winner = msg;
                id = i;
            }
        }
        // send output activation level
        _out.send(winner);
        _outAct.send(winner.act);
        if (winner.act > 0) {
            _outMsg.send(winner.msg);  // plain messages mustn't be sent if activation is 0
            _winner.send(id);  // no one has won if activation is 0
        } else {
            _winner.send(-1);  // no one has won if activation is 0
        }
    }
    
}
