package foraging;

import com.cyberbotics.webots.controller.DistanceSensor;


public class Avoid extends foraging_stub.Avoid_stub {
    
    /* 
     * Khepera's distance sensors layout:
     * 
     *         ds2 ds3
     *      ds1       ds4
     *     ds0         ds5
     * 
     *         ds7 ds6
     */
    
    private final DistanceSensor[] ds;
    private float ds0, ds1, ds2, ds3, ds4, ds5, ds6, ds7;  // [0: no obstacle, 1: obstacle]
    private final FSM fsm;
    
    
    public Avoid(DistanceSensor[] ds) {
        super();
        this.ds = ds;
        fsm = new FSM(IDLE, TURN, AVOID, ESCAPE);
    }
    
    
    @Override
    public void init() {
        fsm.init();
    }
    
    
    @Override
    public void step() {
        // read sensor data
        ds0 = Misc.stdRange((float)ds[0].getValue(), 15, 1000);
        ds1 = Misc.stdRange((float)ds[1].getValue(), 15, 1000);
        ds2 = Misc.stdRange((float)ds[2].getValue(), 15, 1000);
        ds3 = Misc.stdRange((float)ds[3].getValue(), 15, 1000);
        ds4 = Misc.stdRange((float)ds[4].getValue(), 15, 1000);
        ds5 = Misc.stdRange((float)ds[5].getValue(), 15, 1000);
        ds6 = Misc.stdRange((float)ds[6].getValue(), 15, 1000);
        ds7 = Misc.stdRange((float)ds[7].getValue(), 15, 1000);
        fsm.exec();
    }
    
    
    private final FSM.State IDLE = new FSM.State() {
        @Override
        protected void exec() {
            if (ds2 + ds3 > 0) {  // obstacle ahead -> turn
                transition(TURN);
            } else if (ds0 + ds1 + ds4 + ds5 > 0) {  // obstacles on the sides -> avoid
                transition(AVOID);
            } else if (ds6 + ds7 > 0) {  // obstacle behind -> escape
                transition(ESCAPE);
            }
        }
    };
    
    
    private final FSM.State TURN = new FSM.State() {
        private boolean left;
        
        @Override
        protected void entry() {
            // turn towards the less obstructed side
            left = (ds0 + ds1 + ds2 < ds3 + ds4 + ds5);
            // to avoid potential deadlocks sometimes consider the opposite direction
            if (Misc.happens(0.3f)) {
                left = !left;
            }
            _drive.send(new Drive.Speed(left  ?  -1  :  +1,
                                        left  ?  +1  :  -1));
        }
        
        @Override
        protected void exec() {
            if (ds2 + ds3 > 0) {  // obstacle head -> keep turning in the same direction
                _drive.send(new Drive.Speed(left  ?  -1  :  +1,
                                            left  ?  +1  :  -1));
            } else if (ds0 + ds1 + ds4 + ds5 > 0) {  // obstacles on the sides -> avoid
                transition(AVOID);
            } else {  // no more obstacles -> escape
                transition(ESCAPE);
            }
        }
    };
    
    
    private final FSM.State AVOID = new FSM.State() {
        @Override
        protected void entry() {
            // Braitenberg algorithm to avoid side obstacles and follow corridors
            _drive.send(new Drive.Speed(1 + 2*ds0 + 4*ds1 - 4*ds4 - 2*ds5,
                                        1 - 2*ds0 - 4*ds1 + 4*ds4 + 2*ds5));
        }
        
        @Override
        protected void exec() {
            if (ds2 + ds3 > 0) {  // obstacle ahead -> turn
                transition(TURN);
            } else if (ds0 + ds1 + ds4 + ds5 > 0) {  // obstacles on the sides -> keep avoiding
                entry();
            } else {  // no more obstacles -> escape
                transition(ESCAPE);
            }
        }
    };
    
    
    private final FSM.State ESCAPE = new FSM.State() {
        private int wait;
        
        @Override
        protected void entry() {
            wait = Misc.steps(Misc.rndRange(800, 1601));  // [0.8,1.6] seconds
            _drive.send(new Drive.Speed(1, 1));
        }
        
        @Override
        protected void exec() {
            if (ds2 + ds3 > 0) {  // obstacle ahead -> turn
                transition(TURN);
            } else if (ds0 + ds1 + ds4 + ds5 > 0) {  // obstacles on the sides -> avoid
                transition(AVOID);
            } else {
                wait--;
                if (wait <= 0) {  // escape maneuver ended -> reset
                    transition(IDLE);
                } else {
                    _drive.send(new Drive.Speed(1, 1));
                }
            }
        }
    };
    
}
