package foraging;


public class FSM {
    
    public static abstract class State {
        private FSM fsm;
        
        protected void entry() { }
        protected void exec() { }
        protected void exit() { }
        
        protected void transition(State target) {
            fsm.transition(target);
        }
    }
    
    
    private State state;
    private State[] states;
    
    public FSM(State... states) {
        if (states.length < 1) {
            throw new IllegalArgumentException("Each FSM must have at least one state.");
        }
        this.states = states;
        for (State s: states) {
            s.fsm = this;
        }
    }
    
    public State currentState() {
        return state;
    }
    
    public void init() {
        state = states[0];
        state.entry();
    }
    
    public void exec() {
        state.exec();
    }
    
    public void reset() {
        transition(states[0]);
    }
    
    /*public*/ private void transition(State target) {
        if (!contains(target)) {
            throw new IllegalArgumentException("The target state is not part of this FSM.");
        }
        state.exit();
        state = target;
        state.entry();
    }
    
    /*public*/ private boolean contains(State state) {
        for (State s: states) {
            if (s == state) {
                return true;
            }
        }
        return false;
    }
}
